use std::{
    ptr::NonNull,
    sync::{
        Arc, OnceLock, Weak,
        atomic::{AtomicU8, Ordering},
    },
};

use alloc::collections::BTreeMap;

use async_task::Runnable;
use crossbeam_utils::CachePadded;
use parking_lot::{Condvar, Mutex};

use crate::{
    job::{HeapJob, Job, StackJob},
    latch::{AsCoreLatch, HeartbeatLatch, LatchRef, UnsafeWakeLatch},
    workerthread::{HeartbeatThread, WorkerThread},
};

pub struct Heartbeat {
    pub latch: HeartbeatLatch,
}

impl Heartbeat {
    pub fn new() -> NonNull<CachePadded<Self>> {
        let ptr = Box::new(CachePadded::new(Self {
            latch: HeartbeatLatch::new(),
        }));

        Box::into_non_null(ptr)
    }

    pub fn is_pending(&self) -> bool {
        self.latch.as_core_latch().poll_heartbeat()
    }

    pub fn is_sleeping(&self) -> bool {
        self.latch.as_core_latch().is_sleeping()
    }
}

pub struct Context {
    shared: Mutex<Shared>,
    pub shared_job: Condvar,
}

pub(crate) struct Shared {
    pub jobs: BTreeMap<usize, NonNull<Job>>,
    pub heartbeats: BTreeMap<usize, NonNull<CachePadded<Heartbeat>>>,
    injected_jobs: Vec<NonNull<Job>>,
    heartbeat_count: usize,
    should_exit: bool,
}

unsafe impl Send for Shared {}

impl Shared {
    pub fn new_heartbeat(&mut self) -> (NonNull<CachePadded<Heartbeat>>, usize) {
        let index = self.heartbeat_count;
        self.heartbeat_count = index.wrapping_add(1);

        let heatbeat = Heartbeat::new();

        self.heartbeats.insert(index, heatbeat);

        (heatbeat, index)
    }

    pub(crate) fn remove_heartbeat(&mut self, index: usize) {
        self.heartbeats.remove(&index);
    }

    pub fn pop_job(&mut self) -> Option<NonNull<Job>> {
        // this is unlikely, so make the function cold?
        // TODO: profile this
        if !self.injected_jobs.is_empty() {
            // SAFETY: we checked that injected_jobs is not empty
            unsafe { return Some(self.pop_injected_job()) };
        } else {
            self.jobs.pop_first().map(|(_, job)| job)
        }
    }

    #[cold]
    unsafe fn pop_injected_job(&mut self) -> NonNull<Job> {
        self.injected_jobs.pop().unwrap()
    }

    pub fn notify_job_shared(&self) {
        _ = self.heartbeats.iter().find(|(_, heartbeat)| unsafe {
            if heartbeat.as_ref().is_sleeping() {
                heartbeat.as_ref().latch.signal_job_shared();
                return true;
            } else {
                return false;
            }
        });
    }

    pub fn should_exit(&self) -> bool {
        self.should_exit
    }
}

impl Context {
    #[inline]
    pub(crate) fn shared(&self) -> parking_lot::MutexGuard<'_, Shared> {
        self.shared.lock()
    }

    pub fn new_with_threads(num_threads: usize) -> Arc<Self> {
        let this = Arc::new(Self {
            shared: Mutex::new(Shared {
                jobs: BTreeMap::new(),
                heartbeats: BTreeMap::new(),
                injected_jobs: Vec::new(),
                heartbeat_count: 0,
                should_exit: false,
            }),
            shared_job: Condvar::new(),
        });

        tracing::trace!("Creating thread pool with {} threads", num_threads);

        // Create a barrier to synchronize the worker threads and the heartbeat thread
        let barrier = Arc::new(std::sync::Barrier::new(num_threads + 2));

        for i in 0..num_threads {
            let ctx = this.clone();
            let barrier = barrier.clone();

            std::thread::Builder::new()
                .name(format!("worker-{}", i))
                .spawn(move || {
                    let worker = Box::new(WorkerThread::new_in(ctx));

                    barrier.wait();
                    worker.run();
                })
                .expect("Failed to spawn worker thread");
        }

        {
            let ctx = this.clone();
            let barrier = barrier.clone();

            std::thread::Builder::new()
                .name("heartbeat-thread".to_string())
                .spawn(move || {
                    barrier.wait();
                    HeartbeatThread::new(ctx).run();
                })
                .expect("Failed to spawn heartbeat thread");
        }

        barrier.wait();

        this
    }

    pub fn set_should_exit(&self) {
        let mut shared = self.shared.lock();
        shared.should_exit = true;
        for (_, heartbeat) in shared.heartbeats.iter() {
            unsafe {
                heartbeat.as_ref().latch.signal_job_shared();
            }
        }
    }

    pub fn new() -> Arc<Self> {
        Self::new_with_threads(crate::util::available_parallelism())
    }

    pub fn global_context() -> &'static Arc<Self> {
        static GLOBAL_CONTEXT: OnceLock<Arc<Context>> = OnceLock::new();

        GLOBAL_CONTEXT.get_or_init(|| Self::new())
    }

    pub fn inject_job(&self, job: NonNull<Job>) {
        let mut shared = self.shared.lock();
        shared.injected_jobs.push(job);

        shared.notify_job_shared();
    }

    /// Runs closure in this context, processing the other context's worker's jobs while waiting for the result.
    fn run_in_worker_cross<T, F>(self: &Arc<Self>, worker: &WorkerThread, f: F) -> T
    where
        F: FnOnce(&WorkerThread) -> T + Send,
        T: Send,
    {
        // current thread is not in the same context, create a job and inject it into the other thread's context, then wait while working on our jobs.

        // SAFETY: we are waiting on this latch in this thread.
        let latch = unsafe { UnsafeWakeLatch::new(&raw const worker.heartbeat().latch) };

        let job = StackJob::new(
            move || {
                let worker = WorkerThread::current_ref()
                    .expect("WorkerThread::run_in_worker called outside of worker thread");

                f(worker)
            },
            LatchRef::new(&latch),
        );

        let job = job.as_job();
        job.set_pending();

        self.inject_job(Into::into(&job));

        worker.wait_until_latch(&latch);

        let t = unsafe { job.transmute_ref::<T>().wait().into_result() };

        t
    }

    /// Run closure in this context, sleeping until the job is done.
    pub fn run_in_worker_cold<T, F>(self: &Arc<Self>, f: F) -> T
    where
        F: FnOnce(&WorkerThread) -> T + Send,
        T: Send,
    {
        use crate::latch::HeartbeatLatch;
        // current thread isn't a worker thread, create job and inject into global context

        let latch = HeartbeatLatch::new();

        let job = StackJob::new(
            move || {
                let worker = WorkerThread::current_ref()
                    .expect("WorkerThread::run_in_worker called outside of worker thread");

                f(worker)
            },
            LatchRef::new(&latch),
        );

        let job = job.as_job();
        job.set_pending();

        self.inject_job(Into::into(&job));
        latch.wait_and_reset();

        let t = unsafe { job.transmute_ref::<T>().wait().into_result() };

        t
    }

    /// Run closure in this context.
    pub fn run_in_worker<T, F>(self: &Arc<Self>, f: F) -> T
    where
        T: Send,
        F: FnOnce(&WorkerThread) -> T + Send,
    {
        match WorkerThread::current_ref() {
            Some(worker) => {
                // check if worker is in the same context
                if Arc::ptr_eq(&worker.context, self) {
                    tracing::trace!("run_in_worker: current thread");
                    f(worker)
                } else {
                    // current thread is a worker for a different context
                    tracing::trace!("run_in_worker: cross-context");
                    self.run_in_worker_cross(worker, f)
                }
            }
            None => {
                // current thread is not a worker for any context
                tracing::trace!("run_in_worker: inject into context");
                self.run_in_worker_cold(f)
            }
        }
    }
}

impl Context {
    pub fn spawn<F>(self: &Arc<Self>, f: F)
    where
        F: FnOnce() + Send + 'static,
    {
        let job = Box::new(HeapJob::new(f)).into_boxed_job();
        tracing::trace!("Context::spawn: spawning job: {:?}", job);
        unsafe {
            (&*job).set_pending();
            self.inject_job(NonNull::new_unchecked(job));
        }
    }

    pub fn spawn_future<T, F>(self: &Arc<Self>, future: F) -> async_task::Task<T>
    where
        F: Future<Output = T> + Send + 'static,
        T: Send + 'static,
    {
        let schedule = move |runnable: Runnable| {
            #[align(8)]
            unsafe fn harness<T>(this: *const (), job: *const Job<T>) {
                unsafe {
                    let runnable =
                        Runnable::<()>::from_raw(NonNull::new_unchecked(this.cast_mut()));
                    runnable.run();

                    // SAFETY: job was turned into raw
                    drop(Box::from_raw(job.cast_mut()));
                }
            }

            let job = Box::new(Job::<T>::new(harness::<T>, runnable.into_raw()));

            // casting into Job<()> here
            unsafe {
                job.set_pending();
                self.inject_job(NonNull::new_unchecked(Box::into_raw(job) as *mut Job<()>));
            }
        };

        let (runnable, task) = unsafe { async_task::spawn_unchecked(future, schedule) };

        runnable.schedule();

        task
    }

    #[allow(dead_code)]
    fn spawn_async<T, Fut, Fn>(self: &Arc<Self>, f: Fn) -> async_task::Task<T>
    where
        Fn: FnOnce() -> Fut + Send + 'static,
        Fut: Future<Output = T> + Send + 'static,
        T: Send + 'static,
    {
        let future = async move { f().await };

        self.spawn_future(future)
    }
}

pub fn run_in_worker<T, F>(f: F) -> T
where
    T: Send,
    F: FnOnce(&WorkerThread) -> T + Send,
{
    Context::global_context().run_in_worker(f)
}

#[cfg(test)]
mod tests {
    use tracing_test::traced_test;

    use super::*;

    #[test]
    fn run_in_worker_test() {
        let ctx = Context::global_context().clone();
        let result = ctx.run_in_worker(|_| 42);
        assert_eq!(result, 42);
    }

    #[test]
    fn spawn_future_test() {
        let ctx = Context::global_context().clone();
        let task = ctx.spawn_future(async { 42 });

        // Wait for the task to complete
        let result = futures::executor::block_on(task);
        assert_eq!(result, 42);
    }

    #[test]
    fn spawn_async_test() {
        let ctx = Context::global_context().clone();
        let task = ctx.spawn_async(|| async { 42 });

        // Wait for the task to complete
        let result = futures::executor::block_on(task);
        assert_eq!(result, 42);
    }

    #[test]
    fn spawn_test() {
        let ctx = Context::global_context().clone();
        let counter = Arc::new(AtomicU8::new(0));
        let barrier = Arc::new(std::sync::Barrier::new(2));

        ctx.spawn({
            let counter = counter.clone();
            let barrier = barrier.clone();
            move || {
                counter.fetch_add(1, Ordering::SeqCst);
                barrier.wait();
            }
        });

        barrier.wait();
        assert_eq!(counter.load(Ordering::SeqCst), 1);
    }
}
