use core::{
    marker::PhantomData,
    sync::atomic::{AtomicUsize, Ordering},
};
use std::{
    cell::UnsafeCell,
    sync::{Arc, atomic::AtomicU8},
};

use parking_lot::{Condvar, Mutex};

use crate::{WorkerThread, context::Context};

pub trait Latch {
    unsafe fn set_raw(this: *const Self);
}

pub trait Probe {
    fn probe(&self) -> bool;
}

pub type CoreLatch = AtomicLatch;
pub trait AsCoreLatch {
    fn as_core_latch(&self) -> &CoreLatch;
}

#[derive(Debug)]
pub struct AtomicLatch {
    inner: AtomicU8,
}

impl AtomicLatch {
    pub const UNSET: u8 = 0;
    pub const SET: u8 = 1;
    pub const SLEEPING: u8 = 2;
    pub const WAKEUP: u8 = 4;
    pub const HEARTBEAT: u8 = 8;

    #[inline]
    pub const fn new() -> Self {
        Self {
            inner: AtomicU8::new(Self::UNSET),
        }
    }

    pub const fn new_set() -> Self {
        Self {
            inner: AtomicU8::new(Self::SET),
        }
    }

    #[inline]
    pub fn unset(&self) {
        self.inner.fetch_and(!Self::SET, Ordering::Release);
    }

    pub fn reset(&self) -> u8 {
        self.inner.swap(Self::UNSET, Ordering::Release)
    }

    pub fn get(&self) -> u8 {
        self.inner.load(Ordering::Acquire)
    }

    pub fn poll_heartbeat(&self) -> bool {
        self.inner.fetch_and(!Self::HEARTBEAT, Ordering::Relaxed) & Self::HEARTBEAT
            == Self::HEARTBEAT
    }

    /// returns true if the latch was already set.
    pub fn set_sleeping(&self) -> bool {
        self.inner.fetch_or(Self::SLEEPING, Ordering::Relaxed) & Self::SET == Self::SET
    }

    pub fn is_sleeping(&self) -> bool {
        self.inner.load(Ordering::Relaxed) & Self::SLEEPING == Self::SLEEPING
    }

    pub fn is_heartbeat(&self) -> bool {
        self.inner.load(Ordering::Relaxed) & Self::HEARTBEAT == Self::HEARTBEAT
    }

    pub fn is_wakeup(&self) -> bool {
        self.inner.load(Ordering::Relaxed) & Self::WAKEUP == Self::WAKEUP
    }

    pub fn is_set(&self) -> bool {
        self.inner.load(Ordering::Relaxed) & Self::SET == Self::SET
    }

    pub fn set_wakeup(&self) {
        self.inner.fetch_or(Self::WAKEUP, Ordering::Relaxed);
    }

    pub fn set_heartbeat(&self) {
        self.inner.fetch_or(Self::HEARTBEAT, Ordering::Relaxed);
    }

    /// returns true if the latch was previously sleeping.
    #[inline]
    pub unsafe fn set(this: *const Self) -> bool {
        unsafe {
            let old = (*this).inner.fetch_or(Self::SET, Ordering::Relaxed);
            old & Self::SLEEPING == Self::SLEEPING
        }
    }
}

impl Latch for AtomicLatch {
    #[inline]
    unsafe fn set_raw(this: *const Self) {
        unsafe {
            Self::set(this);
        }
    }
}

impl Probe for AtomicLatch {
    #[inline]
    fn probe(&self) -> bool {
        self.inner.load(Ordering::Relaxed) & Self::SET == Self::SET
    }
}
impl AsCoreLatch for AtomicLatch {
    #[inline]
    fn as_core_latch(&self) -> &CoreLatch {
        self
    }
}

pub struct LatchRef<'a, L: Latch> {
    inner: *const L,
    _marker: PhantomData<&'a L>,
}

impl<'a, L: Latch> LatchRef<'a, L> {
    #[inline]
    pub const fn new(latch: &'a L) -> Self {
        Self {
            inner: latch,
            _marker: PhantomData,
        }
    }
}

impl<'a, L: Latch> Latch for LatchRef<'a, L> {
    #[inline]
    unsafe fn set_raw(this: *const Self) {
        unsafe {
            let this = &*this;
            Latch::set_raw(this.inner);
        }
    }
}

impl<'a, L: Latch + Probe> Probe for LatchRef<'a, L> {
    #[inline]
    fn probe(&self) -> bool {
        unsafe {
            let this = &*self.inner;
            Probe::probe(this)
        }
    }
}

impl<'a, L> AsCoreLatch for LatchRef<'a, L>
where
    L: Latch + AsCoreLatch,
{
    #[inline]
    fn as_core_latch(&self) -> &CoreLatch {
        unsafe {
            let this = &*self.inner;
            this.as_core_latch()
        }
    }
}

pub struct NopLatch;

impl Latch for NopLatch {
    #[inline]
    unsafe fn set_raw(_this: *const Self) {
        // do nothing
    }
}

impl Probe for NopLatch {
    #[inline]
    fn probe(&self) -> bool {
        false // always returns false
    }
}

pub struct CountLatch<L: Latch> {
    count: AtomicUsize,
    inner: L,
}

impl<L: Latch> CountLatch<L> {
    #[inline]
    pub const fn new(inner: L) -> Self {
        Self {
            count: AtomicUsize::new(0),
            inner,
        }
    }

    pub fn count(&self) -> usize {
        self.count.load(Ordering::Relaxed)
    }

    pub fn inner(&self) -> &L {
        &self.inner
    }

    #[inline]
    pub fn increment(&self) {
        self.count.fetch_add(1, Ordering::Release);
    }

    #[inline]
    pub fn decrement(&self) {
        unsafe {
            Latch::set_raw(self);
        }
    }
}

impl<L: Latch> Latch for CountLatch<L> {
    #[inline]
    unsafe fn set_raw(this: *const Self) {
        unsafe {
            if (&*this).count.fetch_sub(1, Ordering::Relaxed) == 1 {
                tracing::trace!("CountLatch set_raw: count was 1, setting inner latch");
                // If the count was 1, we need to set the inner latch.
                Latch::set_raw(&(*this).inner);
            }
        }
    }
}

impl<L: Latch + Probe> Probe for CountLatch<L> {
    #[inline]
    fn probe(&self) -> bool {
        self.count.load(Ordering::Relaxed) == 0
    }
}

impl<L: Latch + AsCoreLatch> AsCoreLatch for CountLatch<L> {
    #[inline]
    fn as_core_latch(&self) -> &CoreLatch {
        self.inner.as_core_latch()
    }
}

pub struct HeartbeatLatch {
    inner: UnsafeCell<AtomicLatch>,
    lock: Mutex<()>,
    condvar: Condvar,
}

unsafe impl Send for HeartbeatLatch {}
unsafe impl Sync for HeartbeatLatch {}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub(crate) enum WakeResult {
    Wake,
    Heartbeat,
    Set,
}

impl HeartbeatLatch {
    #[inline]
    pub const fn new() -> Self {
        Self {
            inner: UnsafeCell::new(AtomicLatch::new()),
            lock: Mutex::new(()),
            condvar: Condvar::new(),
        }
    }

    #[inline]
    pub fn reset(&self) {
        let _guard = self.lock.lock();
        // SAFETY: inner is atomic, so we can safely access it.
        unsafe { self.inner.as_mut_unchecked().unset() };
    }

    pub fn wait_and_reset(&self) -> WakeResult {
        // SAFETY: inner is locked by the mutex, so we can safely access it.
        let value = unsafe {
            let mut guard = self.lock.lock();
            let inner = self.inner.as_ref_unchecked();
            inner.set_sleeping();
            while inner.get() & !AtomicLatch::SLEEPING == AtomicLatch::UNSET {
                self.condvar.wait(&mut guard);
            }

            inner.reset()
        };

        if value & AtomicLatch::SET == AtomicLatch::SET {
            WakeResult::Set
        } else if value & AtomicLatch::WAKEUP == AtomicLatch::WAKEUP {
            WakeResult::Wake
        } else if value & AtomicLatch::HEARTBEAT == AtomicLatch::HEARTBEAT {
            WakeResult::Heartbeat
        } else {
            panic!("MutexLatch was not set correctly");
        }
    }

    pub fn set(&self) {
        unsafe {
            Latch::set_raw(self);
        }
    }

    pub fn signal_heartbeat(&self) {
        let mut _guard = self.lock.lock();
        // SAFETY: inner is locked by the mutex, so we can safely access it.
        unsafe {
            let inner = self.inner.as_ref_unchecked();
            inner.set_heartbeat();

            // If the latch was sleeping, notify the waiting thread.
            if inner.is_sleeping() {
                self.condvar.notify_all();
            }
        }
    }

    pub fn signal_job_shared(&self) {
        let mut _guard = self.lock.lock();
        // SAFETY: inner is locked by the mutex, so we can safely access it.
        unsafe {
            self.inner.as_ref_unchecked().set_wakeup();
            if self.inner.as_ref_unchecked().is_sleeping() {
                self.condvar.notify_all();
            }
        }
    }

    pub fn signal_job_finished(&self) {
        let mut _guard = self.lock.lock();
        // SAFETY: inner is locked by the mutex, so we can safely access it.
        unsafe {
            CoreLatch::set(self.inner.get());
            if self.inner.as_ref_unchecked().is_sleeping() {
                self.condvar.notify_all();
            }
        }
    }
}

impl Latch for HeartbeatLatch {
    #[inline]
    unsafe fn set_raw(this: *const Self) {
        // SAFETY: `this` is valid until the guard is dropped.
        unsafe {
            let this = &*this;
            let _guard = this.lock.lock();
            Latch::set_raw(this.inner.get() as *const AtomicLatch);
            if this.inner.as_ref_unchecked().is_sleeping() {
                this.condvar.notify_all();
            }
        }
    }
}

impl Probe for HeartbeatLatch {
    #[inline]
    fn probe(&self) -> bool {
        let _guard = self.lock.lock();
        // SAFETY: inner is atomic, so we can safely access it.
        unsafe { self.inner.as_ref_unchecked().probe() }
    }
}

impl AsCoreLatch for HeartbeatLatch {
    #[inline]
    fn as_core_latch(&self) -> &CoreLatch {
        // SAFETY: inner is atomic, so we can safely access it.
        unsafe { self.inner.as_ref_unchecked() }
    }
}

/// Must only be `set` from a worker thread.
pub struct WakeLatch {
    inner: AtomicLatch,
    worker_index: AtomicUsize,
}

impl WakeLatch {
    pub fn new(worker_index: usize) -> Self {
        Self {
            inner: AtomicLatch::new(),
            worker_index: AtomicUsize::new(worker_index),
        }
    }

    pub(crate) fn set_worker_index(&self, worker_index: usize) {
        self.worker_index.store(worker_index, Ordering::Relaxed);
    }
}

impl Latch for WakeLatch {
    #[inline]
    unsafe fn set_raw(this: *const Self) {
        unsafe {
            let worker_index = (&*this).worker_index.load(Ordering::Relaxed);

            if CoreLatch::set(&(&*this).inner) {
                let ctx = WorkerThread::current_ref().unwrap().context.clone();
                // If the latch was sleeping, wake the worker thread
                ctx.shared()
                    .heartbeats
                    .get(&worker_index)
                    .map(|ptr| ptr.as_ref().latch.signal_job_finished());
            }
        }
    }
}

impl Probe for WakeLatch {
    #[inline]
    fn probe(&self) -> bool {
        self.inner.probe()
    }
}

impl AsCoreLatch for WakeLatch {
    #[inline]
    fn as_core_latch(&self) -> &CoreLatch {
        &self.inner
    }
}

/// A latch that can be set from any thread, but must be created with a valid waker.
pub struct UnsafeWakeLatch {
    waker: *const HeartbeatLatch,
}

impl UnsafeWakeLatch {
    /// # Safety
    /// The `waker` must be valid until the latch is set.
    pub unsafe fn new(waker: *const HeartbeatLatch) -> Self {
        Self { waker }
    }
}

impl Latch for UnsafeWakeLatch {
    #[inline]
    unsafe fn set_raw(this: *const Self) {
        unsafe {
            let waker = (*this).waker;
            Latch::set_raw(waker);
        }
    }
}

impl Probe for UnsafeWakeLatch {
    #[inline]
    fn probe(&self) -> bool {
        // SAFETY: waker is valid as per the constructor contract.
        unsafe {
            let waker = &*self.waker;
            waker.probe()
        }
    }
}

impl AsCoreLatch for UnsafeWakeLatch {
    #[inline]
    fn as_core_latch(&self) -> &CoreLatch {
        // SAFETY: waker is valid as per the constructor contract.
        unsafe {
            let waker = &*self.waker;
            waker.as_core_latch()
        }
    }
}

#[cfg(test)]
mod tests {
    use std::sync::Barrier;

    use tracing::Instrument;
    use tracing_test::traced_test;

    use super::*;

    #[test]
    fn test_atomic_latch() {
        let latch = AtomicLatch::new();
        assert_eq!(latch.get(), AtomicLatch::UNSET);
        unsafe {
            assert!(!latch.probe());
            AtomicLatch::set_raw(&latch);
        }
        assert_eq!(latch.get(), AtomicLatch::SET);
        assert!(latch.probe());
        latch.unset();
        assert_eq!(latch.get(), AtomicLatch::UNSET);
    }

    #[test]
    fn core_latch_sleep() {
        let latch = AtomicLatch::new();
        assert_eq!(latch.get(), AtomicLatch::UNSET);
        latch.set_sleeping();
        assert_eq!(latch.get(), AtomicLatch::SLEEPING);
        unsafe {
            assert!(!latch.probe());
            assert!(AtomicLatch::set(&latch));
        }
        assert_eq!(latch.get(), AtomicLatch::SET | AtomicLatch::SLEEPING);
        assert!(latch.probe());
        latch.reset();
        assert_eq!(latch.get(), AtomicLatch::UNSET);
    }

    #[test]
    fn nop_latch() {
        assert!(
            core::mem::size_of::<NopLatch>() == 0,
            "NopLatch should be zero-sized"
        );
    }

    #[test]
    fn count_latch() {
        let latch = CountLatch::new(AtomicLatch::new());
        assert_eq!(latch.count(), 0);
        latch.increment();
        assert_eq!(latch.count(), 1);
        assert!(!latch.probe());
        latch.increment();
        assert_eq!(latch.count(), 2);
        assert!(!latch.probe());

        unsafe {
            Latch::set_raw(&latch);
        }
        assert!(!latch.probe());
        assert_eq!(latch.count(), 1);

        unsafe {
            Latch::set_raw(&latch);
        }
        assert!(latch.probe());
        assert_eq!(latch.count(), 0);
    }

    #[test]
    #[traced_test]
    fn mutex_latch() {
        let latch = Arc::new(HeartbeatLatch::new());
        assert!(!latch.probe());
        latch.set();
        assert!(latch.probe());
        latch.reset();
        assert!(!latch.probe());

        // Test wait functionality
        let latch_clone = latch.clone();
        let handle = std::thread::spawn(move || {
            assert_eq!(latch_clone.wait_and_reset(), WakeResult::Set);
        });

        // Give the thread time to block
        std::thread::sleep(std::time::Duration::from_millis(100));
        assert!(!latch.probe());

        latch.set();
        handle.join().expect("Thread should join successfully");
    }

    #[test]
    #[traced_test]
    fn wake_latch() {
        let context = Context::new_with_threads(1);
        let count = Arc::new(AtomicUsize::new(0));
        let barrier = Arc::new(Barrier::new(2));

        tracing::info!("running scope in worker thread");
        context.run_in_worker(|worker| {
            tracing::info!("worker thread started: {:?}", worker.index);
            let latch = Arc::new(WakeLatch::new(worker.index));
            worker.context.spawn({
                let heartbeat = unsafe { crate::util::Send::new(worker.heartbeat) };
                let barrier = barrier.clone();
                let count = count.clone();
                let latch = latch.clone();
                move || {
                    tracing::info!("sleeping workerthread");

                    latch.as_core_latch().set_sleeping();
                    unsafe {
                        heartbeat.as_ref().latch.wait_and_reset();
                    }
                    tracing::info!("woken up workerthread");
                    count.fetch_add(1, Ordering::SeqCst);
                    tracing::info!("waiting on barrier");
                    barrier.wait();
                }
            });

            worker.context.spawn({
                move || {
                    tracing::info!("setting latch in worker thread");
                    unsafe {
                        Latch::set_raw(&*latch);
                    }
                }
            });
        });

        tracing::info!("main thread set latch, waiting for worker thread to wake up");
        barrier.wait();
        assert_eq!(
            count.load(Ordering::SeqCst),
            1,
            "Latch should have woken the worker thread"
        );
    }
}
