use std::{
    cell::{Cell, UnsafeCell},
    ptr::NonNull,
    sync::Arc,
    time::Duration,
};

use crossbeam_utils::CachePadded;

use crate::{
    context::{Context, Heartbeat},
    job::{Job, JobList, JobResult},
    latch::{AsCoreLatch, CoreLatch, Probe},
    util::DropGuard,
};

pub struct WorkerThread {
    pub(crate) context: Arc<Context>,
    pub(crate) index: usize,
    pub(crate) queue: UnsafeCell<JobList>,
    pub(crate) heartbeat: NonNull<CachePadded<Heartbeat>>,
    pub(crate) join_count: Cell<u8>,
}

impl Drop for WorkerThread {
    fn drop(&mut self) {
        // remove the current worker thread from the heartbeat list
        self.context.shared().remove_heartbeat(self.index);

        // SAFETY: we removed the heartbeat from the context, so we can safely drop it.
        unsafe {
            _ = Box::from_non_null(self.heartbeat);
        }
    }
}

thread_local! {
    static WORKER: UnsafeCell<Option<NonNull<WorkerThread>>> = const { UnsafeCell::new(None) };
}

impl WorkerThread {
    pub fn new_in(context: Arc<Context>) -> Self {
        let (heartbeat, index) = context.shared().new_heartbeat();

        Self {
            context,
            index,
            queue: UnsafeCell::new(JobList::new()),
            heartbeat,
            join_count: Cell::new(0),
        }
    }

    pub(crate) fn heartbeat(&self) -> &CachePadded<Heartbeat> {
        // SAFETY: the heartbeat is always set when the worker thread is created
        unsafe { self.heartbeat.as_ref() }
    }
}

impl WorkerThread {
    pub fn run(self: Box<Self>) {
        let this = Box::into_raw(self);
        unsafe {
            Self::set_current(this);
        }

        let _guard = DropGuard::new(|| unsafe {
            // SAFETY: this is only called when the thread is exiting
            Self::unset_current();
            Self::drop_in_place(this);
        });

        tracing::trace!("WorkerThread::run: starting worker thread");

        unsafe {
            (&*this).run_inner();
        }

        tracing::trace!("WorkerThread::run: worker thread finished");
    }

    fn run_inner(&self) {
        let mut job = self.context.shared().pop_job();
        'outer: loop {
            while let Some(j) = job {
                self.execute(j);

                let mut guard = self.context.shared();
                if guard.should_exit() {
                    // if the context is stopped, break out of the outer loop which
                    // will exit the thread.
                    break 'outer;
                }

                // we executed the shared job, now we want to check for any
                // local jobs which this job might have spawned.
                job = self.pop_front().or_else(|| guard.pop_job());
            }

            // no more jobs, wait to be notified of a new job or a heartbeat.
            match self.heartbeat().latch.wait_and_reset() {
                crate::latch::WakeResult::Wake => {
                    let mut guard = self.context.shared();
                    if guard.should_exit() {
                        break 'outer;
                    }

                    job = guard.pop_job();
                }
                crate::latch::WakeResult::Heartbeat => {
                    self.tick();
                }
                crate::latch::WakeResult::Set => {
                    // check if we should exit the thread
                    if self.context.shared().should_exit() {
                        break 'outer;
                    }
                    panic!("this thread shouldn't be woken by a finished job")
                }
            }
        }
    }
}

impl WorkerThread {
    #[inline(always)]
    pub(crate) fn tick(&self) {
        if self.heartbeat().is_pending() {
            tracing::trace!("received heartbeat, thread id: {:?}", self.index);
            self.heartbeat_cold();
        }
    }

    #[inline]
    fn execute(&self, job: NonNull<Job>) {
        self.tick();
        Job::execute(job);
    }

    #[cold]
    fn heartbeat_cold(&self) {
        let mut guard = self.context.shared();

        if !guard.jobs.contains_key(&self.index) {
            if let Some(job) = self.pop_back() {
                tracing::trace!("heartbeat: sharing job: {:?}", job);
                unsafe {
                    job.as_ref().set_pending();
                }
                guard.jobs.insert(self.index, job);
                guard.notify_job_shared();
            }
        }
    }
}

impl WorkerThread {
    #[inline]
    pub fn pop_back(&self) -> Option<NonNull<Job>> {
        unsafe { self.queue.as_mut_unchecked().pop_back() }
    }

    #[inline]
    pub fn push_back(&self, job: *const Job) {
        unsafe { self.queue.as_mut_unchecked().push_back(job) }
    }

    #[inline]
    pub fn pop_front(&self) -> Option<NonNull<Job>> {
        unsafe { self.queue.as_mut_unchecked().pop_front() }
    }

    #[inline]
    pub fn push_front(&self, job: *const Job) {
        unsafe { self.queue.as_mut_unchecked().push_front(job) }
    }
}

impl WorkerThread {
    #[inline]
    pub fn current_ref<'a>() -> Option<&'a Self> {
        unsafe { (*WORKER.with(UnsafeCell::get)).map(|ptr| ptr.as_ref()) }
    }

    unsafe fn set_current(this: *const Self) {
        WORKER.with(|cell| {
            unsafe {
                // SAFETY: this cell is only ever accessed from the current thread
                assert!(
                    (&mut *cell.get())
                        .replace(NonNull::new_unchecked(
                            this as *const WorkerThread as *mut WorkerThread,
                        ))
                        .is_none()
                );
            }
        });
    }

    unsafe fn unset_current() {
        WORKER.with(|cell| {
            unsafe {
                // SAFETY: this cell is only ever accessed from the current thread
                (&mut *cell.get()).take();
            }
        });
    }

    unsafe fn drop_in_place(this: *mut Self) {
        unsafe {
            // SAFETY: this is only called when the thread is exiting, so we can
            // safely drop the thread. We use `drop_in_place` to prevent `Box`
            // from creating a no-alias reference to the worker thread.
            core::ptr::drop_in_place(this);
            _ = Box::<core::mem::ManuallyDrop<Self>>::from_raw(this as _);
        }
    }
}

pub struct HeartbeatThread {
    ctx: Arc<Context>,
}

impl HeartbeatThread {
    const HEARTBEAT_INTERVAL: Duration = Duration::from_micros(100);

    pub fn new(ctx: Arc<Context>) -> Self {
        Self { ctx }
    }

    pub fn run(self) {
        tracing::trace!("new heartbeat thread {:?}", std::thread::current());

        let mut i = 0;
        loop {
            let sleep_for = {
                let guard = self.ctx.shared();
                if guard.should_exit() {
                    break;
                }

                if let Some((_, heartbeat)) = guard.heartbeats.iter().nth(i) {
                    unsafe {
                        heartbeat.as_ref().latch.signal_heartbeat();
                    }
                }
                let num_heartbeats = guard.heartbeats.len();

                drop(guard);

                if i >= num_heartbeats {
                    i = 0;
                } else {
                    i += 1;
                }

                Self::HEARTBEAT_INTERVAL.checked_div(num_heartbeats as u32)
            };

            if let Some(duration) = sleep_for {
                std::thread::sleep(duration);
            }
        }
    }
}

impl WorkerThread {
    #[cold]
    fn wait_until_latch_cold(&self, latch: &CoreLatch) {
        'outer: while !latch.probe() {
            // process local jobs before locking shared context
            while let Some(job) = self.pop_front() {
                tracing::trace!("thread {:?} executing local job: {:?}", self.index, job);
                unsafe {
                    job.as_ref().set_pending();
                }
                Job::execute(job);
                tracing::trace!("thread {:?} finished local job: {:?}", self.index, job);
            }

            // take a shared job, if it exists
            'inner: loop {
                if let Some(shared_job) = self.context.shared().jobs.remove(&self.index) {
                    tracing::trace!(
                        "thread {:?} executing shared job: {:?}",
                        self.index,
                        shared_job
                    );
                    Job::execute(shared_job);
                }

                while !latch.probe() {
                    tracing::trace!("thread {:?} looking for shared jobs", self.index);

                    let job = {
                        let mut guard = self.context.shared();
                        guard.jobs.remove(&self.index).or_else(|| guard.pop_job())
                    };

                    match job {
                        Some(job) => {
                            tracing::trace!("thread {:?} found job: {:?}", self.index, job);
                            Job::execute(job);

                            continue 'outer;
                        }
                        None => {
                            // TODO: wait on latch? if we have something that can
                            // signal being done, e.g. can be waited on instead of
                            // shared jobs, we should wait on it instead, but we
                            // would also want to receive shared jobs still?
                            // Spin? probably just wastes CPU time.
                            // self.context.shared_job.wait(&mut guard);
                            // if spin.spin() {
                            //     // wait for more shared jobs.
                            //     // self.context.shared_job.wait(&mut guard);
                            //     return;
                            // }
                            // Yield? same as spinning, really, so just exit and let the upstream use wait
                            // std::thread::yield_now();

                            tracing::trace!("thread {:?} is sleeping", self.index);

                            match self.heartbeat().latch.wait_and_reset() {
                                // why were we woken up?
                                // 1. the heartbeat thread ticked and set the
                                // latch, so we should see if we have any work
                                // to share.
                                // 2. a job was shared and we were notified, so
                                // we should execute it.
                                // 3. the job we were waiting on was completed,
                                // so we should return it.
                                crate::latch::WakeResult::Set => {
                                    break 'outer; // we were woken up by a job being set, so we should exit the loop.
                                }
                                crate::latch::WakeResult::Wake => {
                                    // skip checking for local jobs, since we
                                    // were woken up to check for shared jobs.
                                    continue 'inner;
                                }
                                crate::latch::WakeResult::Heartbeat => {
                                    self.tick();
                                    continue 'outer;
                                }
                            }
                            // since we were sleeping, the shared job can't be populated,
                            // so resuming the inner loop is fine.
                        }
                    }
                }

                break;
            }
        }

        tracing::trace!(
            "thread {:?} finished waiting on latch {:?}",
            self.index,
            latch
        );
        self.heartbeat().latch.as_core_latch().unset();
        return;
    }

    pub fn wait_until_job<T>(&self, job: &Job<T>, latch: &CoreLatch) -> Option<JobResult<T>> {
        // we've already checked that the job was popped from the queue
        // check if shared job is our job
        if let Some(shared_job) = self.context.shared().jobs.remove(&self.index) {
            if core::ptr::eq(shared_job.as_ptr(), job as *const Job<T> as _) {
                // this is the job we are looking for, so we want to
                // short-circuit and call it inline
                return None;
            } else {
                // this isn't the job we are looking for, but we still need to
                // execute it
                Job::execute(shared_job);
            }
        }

        // do the usual thing and wait for the job's latch
        if !latch.probe() {
            self.wait_until_latch_cold(latch);
        }

        Some(job.wait())
    }

    pub fn wait_until_latch<L>(&self, latch: &L)
    where
        L: AsCoreLatch,
    {
        let latch = latch.as_core_latch();
        if !latch.probe() {
            tracing::trace!("thread {:?} waiting on latch {:?}", self.index, latch);
            self.wait_until_latch_cold(latch)
        }
    }
}
